/* GTK - The GIMP Toolkit
 * Copyright (C) 2002 Choe Hwanjin
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Author: Choe Hwanjin <krisna@kldp.org>
 */
#include <string.h>

#include <gtk/gtk.h>
#include <gtk/gtkimcontext.h>
#include <gdk/gdkkeysyms.h>

#include "gtkimcontexthangul.h"

static void gtk_im_context_hangul_class_init (GtkIMContextHangulClass *class);
static void gtk_im_context_hangul_init (GtkIMContextHangul *context_hangul);
static void gtk_im_context_hangul_finalize(GObject *obj);

static gboolean	gtk_im_context_hangul_filter_keypress(GtkIMContext *context,
						      GdkEventKey *key);
static void gtk_im_context_hangul_get_preedit_string(GtkIMContext *ic,
						     gchar **str,
						     PangoAttrList **attrs,
						     gint *cursor_pos);

static GObjectClass *parent_class;
GType gtk_type_im_context_hangul = 0;

static guint16 gtk_compose_ignore[] = {
  GDK_Shift_L,
  GDK_Shift_R,
  GDK_Control_L,
  GDK_Control_R,
  GDK_Caps_Lock,
  GDK_Shift_Lock,
  GDK_Meta_L,
  GDK_Meta_R,
  GDK_Alt_L,
  GDK_Alt_R,
  GDK_Super_L,
  GDK_Super_R,
  GDK_Hyper_L,
  GDK_Hyper_R,
  GDK_Mode_switch,
  0
};

void
gtk_im_context_hangul_register_type(GTypeModule *type_module,
				    const gchar *type_name)
{
  static const GTypeInfo im_context_hangul_info = {
    sizeof(GtkIMContextHangulClass),
    (GBaseInitFunc) NULL,
    (GBaseFinalizeFunc) NULL,
    (GClassInitFunc) gtk_im_context_hangul_class_init,
    NULL,
    NULL,
    sizeof(GtkIMContextHangul),
    0,
    (GtkObjectInitFunc) gtk_im_context_hangul_init,
  };

  gtk_type_im_context_hangul =
      g_type_module_register_type(type_module,
				  GTK_TYPE_IM_CONTEXT,
				  type_name,
				  &im_context_hangul_info, 0);
}

static void 
gtk_im_context_hangul_class_init (GtkIMContextHangulClass *class)
{
  GtkIMContextClass *im_context_class = GTK_IM_CONTEXT_CLASS(class);
  GObjectClass *gobject_class = G_OBJECT_CLASS(class);

  parent_class = g_type_class_peek_parent (class);

  im_context_class->filter_keypress = gtk_im_context_hangul_filter_keypress;
  im_context_class->get_preedit_string = gtk_im_context_hangul_get_preedit_string;
  gobject_class->finalize = gtk_im_context_hangul_finalize;
}

static void 
gtk_im_context_hangul_init (GtkIMContextHangul *context_hangul)
{
  context_hangul->state = 0;
  context_hangul->stack[0] = 0;
  context_hangul->stack[1] = 0;
  context_hangul->stack[2] = 0;
  context_hangul->stack[3] = 0;
  context_hangul->stack[4] = 0;
  context_hangul->stack[5] = 0;
  context_hangul->stack[6] = 0;
  context_hangul->stack[7] = 0;
  context_hangul->stack[8] = 0;
  context_hangul->stack[9] = 0;
  context_hangul->index = -1;

  context_hangul->choseong = 0;
  context_hangul->jungseong = 0;
  context_hangul->jongseong = 0;
  context_hangul->automata = NULL;	/* initial value: automata == null */
}

static void
gtk_im_context_hangul_finalize(GObject *obj)
{
  parent_class->finalize (obj);
}

GtkIMContext *
gtk_im_context_hangul_new(void)
{
  return GTK_IM_CONTEXT (g_object_new (GTK_TYPE_IM_CONTEXT_HANGUL, NULL));
}

void
gtk_im_context_hangul_push(GtkIMContextHangul *context_hangul,
			   gunichar ch)
{
  g_print("push: %d U+%04x\n", context_hangul->index + 1, ch);
  context_hangul->stack[++context_hangul->index] = ch;
}

gunichar
gtk_im_context_hangul_peek(GtkIMContextHangul *context_hangul)
{
  if (context_hangul->index < 0)
    return 0;
  return context_hangul->stack[context_hangul->index];
}

gunichar
gtk_im_context_hangul_pop(GtkIMContextHangul *context_hangul)
{
  g_print("pop: %d U+%04x\n", context_hangul->index, context_hangul->stack[context_hangul->index]);
  if (context_hangul->index < 0)
    return 0;
  return context_hangul->stack[context_hangul->index--];
}

void
gtk_im_context_hangul_set_automata(GtkIMContextHangul *context_hangul,
				   GtkIMContextHangulAutomata automata)
{
  context_hangul->automata = automata;
}

#define HCF	0x115f
#define HJF	0x1160

static int
gtk_im_context_hangul_preedit_string(GtkIMContextHangul *context_hangul,
				     gchar *buf)
{
  int n = 0;

  if (context_hangul->choseong == 0 && context_hangul->jungseong == 0 &&
      context_hangul->jongseong == 0) {
    buf[0] = '\0';
    /* g_print("PreEdit: (%d)\n", 0); */
    return 0;
  }
  
  if (context_hangul->choseong == 0)
    n = g_unichar_to_utf8(HCF, buf);
  else
    n = g_unichar_to_utf8(context_hangul->choseong, buf);
  buf[n] = '\0';

  if (context_hangul->jungseong == 0)
    n += g_unichar_to_utf8(HJF, buf + n);
  else
    n += g_unichar_to_utf8(context_hangul->jungseong, buf + n);
  buf[n] = '\0';

  if (context_hangul->jongseong != 0) {
    n += g_unichar_to_utf8(context_hangul->jongseong, buf + n);
    buf[n] = '\0';
  }

  /*
  g_print("PreEdit: U+%04x U+%04x U+%04x (%d)\n", context_hangul->choseong,
		 			    context_hangul->jungseong,
					    context_hangul->jongseong, n);
  */
  return n;
}

static void
gtk_im_context_hangul_get_preedit_string(GtkIMContext *ic, gchar **str,
			     PangoAttrList **attrs,
			     gint *cursor_pos)
{
  gchar buf[24];
  int len = 0;
  GtkIMContextHangul *context_hangul = GTK_IM_CONTEXT_HANGUL(ic);

  buf[0] = '\0';
  len = gtk_im_context_hangul_preedit_string(context_hangul, buf);

  if (attrs) {
    PangoAttribute *attr = pango_attr_underline_new(PANGO_UNDERLINE_SINGLE);
    *attrs = pango_attr_list_new();
    attr->start_index = 0;
    attr->end_index = len;
    pango_attr_list_insert (*attrs, attr);
  }
  if (cursor_pos) {
    *cursor_pos = len;
  }
  if (str) {
    *str = g_strdup(buf);
  }
}

static gboolean
gtk_im_context_hangul_is_ignore_key(guint16 key)
{
  int i;
  for (i = 0; gtk_compose_ignore[i]; i++) {
    if (key == gtk_compose_ignore[i]) {
      return TRUE;
    }
  }
  return FALSE;
}

gboolean
gtk_im_context_hangul_is_trigger(GdkEventKey *key)
{
  return (key->keyval == GDK_space && (key->state & GDK_SHIFT_MASK));
}

gboolean
gtk_im_context_hangul_is_backspace(GdkEventKey *key)
{
  return (key->keyval == GDK_BackSpace);
}

gunichar gtk_im_context_hangul_compchoseong_to_single(gunichar ch)
{
  switch (ch) {
    case 0x1101:	/* hangul choseong ssangkiyeok */
      return 0x1100;	/* hangul choseong kiyeok */
    case 0x1104:	/* hangul choseong ssangtikeut */
      return 0x1103;	/* hangul choseong tikeut */
    case 0x1108:	/* hangul choseong ssangpieup */
      return 0x1107;	/* hangul choseong pieup */
    case 0x110a:	/* hangul choseong ssangsios */
      return 0x1109;	/* hangul choseong sios */
    case 0x110d:	/* hangul choseong ssangcieuc */
      return 0x110c;	/* hangul choseong cieuc */
  }
  return 0;
}

gunichar gtk_im_context_hangul_compjungseong_to_single(gunichar ch)
{
  switch (ch) {
    case 0x116a:	/* hangul jungseong wa */
    case 0x116b:	/* hangul jungseong wae */
    case 0x116c:	/* hangul jungseong oe */
      return 0x1169;	/* hangul jungseong o */
    case 0x116f:	/* hangul jungseong weo */
    case 0x1170:	/* hangul jungseong we */
    case 0x1171:	/* hangul jungseong wi */
      return 0x116e;	/* hangul jungseong u */
    case 0x1174:	/* hangul jungseong yi */
      return 0x1173;	/* hangul jungseong eu */
  }
  return 0;
}

gunichar gtk_im_context_hangul_compjongseong_to_single(gunichar ch)
{
  switch (ch) {
    case 0x11a9:	/* hangul jongseong ssangkiyeok */
    case 0x11aa:	/* hangul jongseong kiyeok-sios */
      return 0x11a8;	/* hangul jongseong kiyeok */
    case 0x11ac:	/* hangul jongseong nieun-cieuc */
    case 0x11ad:	/* hangul jongseong nieun-hieuh */
      return 0x11ab;	/* hangul jongseong nieun */
    case 0x11b0:	/* hangul jongseong rieul-kiyeok */
    case 0x11b1:	/* hangul jongseong rieul-mieum */
    case 0x11b2:	/* hangul jongseong rieul-pieup */
    case 0x11b3:	/* hangul jongseong rieul-sios */
    case 0x11b4:	/* hangul jongseong rieul-thieuth */
    case 0x11b5:	/* hangul jongseong rieul-phieuph */
    case 0x11b6:	/* hangul jongseong rieul-hieuh */
      return 0x11af;	/* hangul jongseong rieul */
    case 0x11b9:	/* hangul jongseong ssangsios */
      return 0x11b8;	/* hangul jongseong sios */
  }
  return 0;
}

gunichar gtk_im_context_hangul_choseong_to_unicode(gunichar ch) {
  static gunichar table[] = {
    0x3131,	/* 0x1100 */
    0x3132,	/* 0x1101 */
    0x3134,	/* 0x1102 */
    0x3137,	/* 0x1103 */
    0x3138,	/* 0x1104 */
    0x3139,	/* 0x1105 */
    0x3141,	/* 0x1106 */
    0x3142,	/* 0x1107 */
    0x3143,	/* 0x1108 */
    0x3145,	/* 0x1109 */
    0x3146,	/* 0x110a */
    0x3147,	/* 0x110b */
    0x3148,	/* 0x110c */
    0x3149,	/* 0x110d */
    0x314a,	/* 0x110e */
    0x314b,	/* 0x110f */
    0x314c,	/* 0x1110 */
    0x314d,	/* 0x1111 */
    0x314e	/* 0x1112 */
  };
  if (!gtk_im_context_hangul_is_choseong(ch))
    return 0;
  return table[ch - 0x1100];
}

gunichar gtk_im_context_hangul_jungseong_to_unicode(gunichar ch) {
  static gunichar table[] = {
    0x314f,	/* 0x1161 */
    0x3150,	/* 0x1162 */
    0x3151,	/* 0x1163 */
    0x3152,	/* 0x1164 */
    0x3153,	/* 0x1165 */
    0x3154,	/* 0x1166 */
    0x3155,	/* 0x1167 */
    0x3156,	/* 0x1168 */
    0x3157,	/* 0x1169 */
    0x3158,	/* 0x116a */
    0x3159,	/* 0x116b */
    0x315a,	/* 0x116c */
    0x315b,	/* 0x116d */
    0x315c,	/* 0x116e */
    0x315d,	/* 0x116f */
    0x315e,	/* 0x1170 */
    0x315f,	/* 0x1171 */
    0x3160,	/* 0x1172 */
    0x3161,	/* 0x1173 */
    0x3162,	/* 0x1174 */
    0x3163	/* 0x1175 */
  };

  if (!gtk_im_context_hangul_is_jungseong(ch))
    return 0;
  return table[ch - 0x1161];
}

gunichar gtk_im_context_hangul_jongseong_to_unicode(gunichar ch) {
  static gunichar table[] = {
    0x3131,	/* 0x11a8 */
    0x3132,	/* 0x11a9 */
    0x3133,	/* 0x11aa */
    0x3134,	/* 0x11ab */
    0x3135,	/* 0x11ac */
    0x3136,	/* 0x11ad */
    0x3137,	/* 0x11ae */
    0x3139,	/* 0x11af */
    0x313a,	/* 0x11b0 */
    0x313b,	/* 0x11b1 */
    0x313c,	/* 0x11b2 */
    0x313d,	/* 0x11b3 */
    0x313e,	/* 0x11b4 */
    0x313f,	/* 0x11b5 */
    0x3140,	/* 0x11b6 */
    0x3141,	/* 0x11b7 */
    0x3142,	/* 0x11b8 */
    0x3144,	/* 0x11b9 */
    0x3145,	/* 0x11ba */
    0x3146,	/* 0x11bb */
    0x3147,	/* 0x11bc */
    0x3148,	/* 0x11bd */
    0x314a,	/* 0x11be */
    0x314b,	/* 0x11bf */
    0x314c,	/* 0x11c0 */
    0x314d,	/* 0x11c1 */
    0x314e	/* 0x11c2 */
  };
  if (!gtk_im_context_hangul_is_jongseong(ch))
    return 0;
  return table[ch - 0x11a8];
}

gunichar
gtk_im_context_hangul_jamo_to_unicode	(gunichar choseong,
				gunichar jungseong,
				gunichar jongseong)
{
  static gunichar hangul_base = 0xac00;
  static gunichar choseong_base = 0x1100;
  static gunichar jungseong_base = 0x1161;
  static gunichar jongseong_base = 0x11a7;
/*  static int nchoseong = 19; */
  static int njungseong = 21;
  static int njongseong = 28;
/*  static int njungjong = 21 * 28; */
  gunichar ch;

  if (!gtk_im_context_hangul_is_choseong(choseong))
    return 0;
  if (!gtk_im_context_hangul_is_jungseong(jungseong))
    return 0;
  if (!gtk_im_context_hangul_is_jongseong(jongseong))
    return 0;

  choseong -= choseong_base;
  jungseong -= jungseong_base;
  jongseong -= jongseong_base;
 
  ch = ((choseong * njungseong) + jungseong) * njongseong + jongseong
	 + hangul_base;
  if (ch >= 0xAC00 && ch <= 0xD7AF)  /* check hangul code */
    return ch;
  else
    return 0;
}

static void gtk_im_context_hangul_reset(GtkIMContextHangul *context_hangul)
{
  context_hangul->index = -1;
  context_hangul->stack[0] = 0;
  context_hangul->stack[1] = 0;
  context_hangul->stack[2] = 0;
  context_hangul->stack[3] = 0;
  context_hangul->stack[4] = 0;
  context_hangul->stack[5] = 0;
  context_hangul->stack[6] = 0;
  context_hangul->stack[7] = 0;
  context_hangul->stack[8] = 0;
  context_hangul->stack[9] = 0;
  context_hangul->choseong = 0;
  context_hangul->jungseong = 0;
  context_hangul->jongseong = 0;
  context_hangul->state = 0;
}

gboolean
gtk_im_context_hangul_process_nonhangul(GtkIMContextHangul *context_hangul,
				        GdkEventKey *key)
{
  if (!gtk_im_context_hangul_is_modifier(key)) {
    gunichar ch = gdk_keyval_to_unicode(key->keyval);
    if (ch != 0) {
      gchar buf[10];
      gint n = g_unichar_to_utf8(ch, buf);
      buf[n] = '\0';
      /* g_print("Nonhangul: %s(U+%04x)\n", buf, ch); */
      g_signal_emit_by_name(context_hangul, "commit", buf);
      return TRUE;
    } 
  }
  return FALSE;
}

static gboolean
gtk_im_context_hangul_handle_direct_mode(GtkIMContextHangul *context_hangul,
				         GdkEventKey *key)
{
  if (gtk_im_context_hangul_is_trigger(key)) {
    context_hangul->state = STATE_HANGUL;
    gtk_im_context_hangul_reset(context_hangul);
    g_signal_emit_by_name (context_hangul, "preedit_changed");
    return TRUE;
  }
  return gtk_im_context_hangul_process_nonhangul(context_hangul, key);
}

void
gtk_im_context_hangul_commit(GtkIMContextHangul *context_hangul)
{
  int n = 0;
  gchar buf[12];
  gunichar ch;
 
  ch = gtk_im_context_hangul_jamo_to_unicode(context_hangul->choseong,
					     context_hangul->jungseong,
	(context_hangul->jongseong)? context_hangul->jongseong : 0x11a7);

  buf[0] = '\0';
  if (ch) {
    n = g_unichar_to_utf8(ch, buf);
    buf[n] = '\0';
    /* g_print("[commit char]: %s(U+%04x)\n", buf, ch); */
  } else {
    if (context_hangul->choseong) {
      ch = gtk_im_context_hangul_choseong_to_unicode(context_hangul->choseong);
      n = g_unichar_to_utf8(ch, buf);
      buf[n] = '\0';
      /* g_print("[commit char]: %s(U+%x)\n", buf, ch); */
    }
    if (context_hangul->jungseong) {
      ch = gtk_im_context_hangul_jungseong_to_unicode(context_hangul->jungseong);
      n += g_unichar_to_utf8(ch, buf + n);
      buf[n] = '\0';
      /* g_print("[commit char]: %s(U+%x)\n", buf, ch); */
    }
    if (context_hangul->jongseong) {
      ch = gtk_im_context_hangul_jongseong_to_unicode(context_hangul->jongseong);
      n += g_unichar_to_utf8(ch, buf + n);
      buf[n] = '\0';
      /* g_print("[commit char]: %s(U+%x)\n", buf, ch); */
    }
  }

  context_hangul->choseong = 0;
  context_hangul->jungseong = 0;
  context_hangul->jongseong = 0;
  context_hangul->index = -1;

  g_signal_emit_by_name(context_hangul, "commit", buf);
}

void
gtk_im_context_hangul_commit_unicode(GtkIMContextHangul *context_hangul,
			     gunichar ch)
{
  int n;
  gchar buf[6];

  n = g_unichar_to_utf8(ch, buf);
  buf[n] = '\0';

  context_hangul->choseong = 0;
  context_hangul->jungseong = 0;
  context_hangul->jongseong = 0;
  context_hangul->index = -1;

  /* g_print("[commit char]: %s(U+%x)\n", buf, ch); */
  g_signal_emit_by_name(context_hangul, "commit", buf);
}

void 
gtk_im_context_hangul_commit_choseong(GtkIMContextHangul *context_hangul,
			      gunichar choseong)
{
  gunichar ch = gtk_im_context_hangul_choseong_to_unicode(choseong);
  gtk_im_context_hangul_commit_unicode(context_hangul, ch);
  context_hangul->choseong = 0;
}

void
gtk_im_context_hangul_commit_jungseong(GtkIMContextHangul *context_hangul,
			       gunichar jungseong)
{
  gunichar ch = gtk_im_context_hangul_jungseong_to_unicode(jungseong);
  gtk_im_context_hangul_commit_unicode(context_hangul, ch);
  context_hangul->jungseong = 0;
}

void 
gtk_im_context_hangul_commit_jongseong(GtkIMContextHangul *context_hangul,
			      gunichar jongseong)
{
  gunichar ch = gtk_im_context_hangul_jongseong_to_unicode(jongseong);
  gtk_im_context_hangul_commit_unicode(context_hangul, ch);
  context_hangul->jongseong = 0;
}

/* use hangul automata */
static gboolean
gtk_im_context_hangul_filter_keypress(GtkIMContext *context,
				      GdkEventKey *key)
{
  GtkIMContextHangul *context_hangul = GTK_IM_CONTEXT_HANGUL(context);

  /* ignore key release */
  if (key->type == GDK_KEY_RELEASE) {
    return FALSE;
  }
  /* some keys are ignored */
  if (gtk_im_context_hangul_is_ignore_key(key->keyval)) {
    return FALSE;
  }
  /* handle direct mode */
  if (context_hangul->state == STATE_DIRECT) {
    return gtk_im_context_hangul_handle_direct_mode(context_hangul, key);
  }

  if (context_hangul->automata)
    return context_hangul->automata(context_hangul, key);
  else
    return FALSE;
}

/* vim: nocindent
 */
